#ifndef _COMMON_H_
#define _COMMON_H_
/*  common.h - some common constants and definitions for lincardemu
    Copyright (C) 2000  The Linux Card Emulator Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    http://www.fsf.org/
*/

#include "debug.h"

#define COMMAND_OK  0x9000
#define COMMAND_KO  0x9008
#define HASH_KO     0x9040

#define F_EURO_S2       0
#define F_TRIPLE_DES    1

#define SetBit(addr, bit) (addr) |= (1 << (bit))
#define UnsetBit(addr, bit) (addr) &= ~(1<< (bit))
#define TestBit(addr, bit) ((addr) & (1 << bit))

typedef unsigned char byte;
typedef unsigned char u8;
typedef signed char s8;
typedef unsigned short u16;
typedef signed short s16;
typedef unsigned int  u32;
typedef signed int  s32;

enum {  DATE_UNDEFINED=-1 };

typedef struct datetime {
  s16 year;
  s8  month;
  s8  mday;
  s8  hour;
  s8  min;
  s8  sec;
} datetime;

#ifdef _WIN32
#ifndef __CYGWIN__
#define USE_WIN32
#endif
#endif

#ifdef _WIN32
  #include "SerialWin.h"
#else
  #include "serfunc.h"
#endif

enum { Sys_Eurocrypt=0, Sys_Viaccess=1, Sys_Conax=2, Sys_Seca=3,
       Sys_Irdeto=4, Sys_Videoguard=5, Sys_EC_S=6, Sys_Systems=7 };
enum { Sys_Betacrypt = Sys_Systems,
       Sys_Nagra, Sys_TPSCrypt, Sys_Skycript, Sys_Cryptoworks, Sys_ExtraSystems,
       FILTER_LENGTH=8 };
enum { Old_Filter_File=1 };

/* for ec_keys.h and seca_keys.h  : */
enum { DECODE_LOCAL=1, DECODE_CWS=2, DECODE_LOCAL_CWS=6, DECODE_CWS_LOCAL=9 ,
       DECODE_ECM_MASK=0xf };

extern char *data_dir;
extern char *log_dir;

extern short c_system;
extern int  auto_save, auto_check;

extern char *decoder_name;

enum { LOGMASKTXT=36, LOGMASKN=5 };
typedef struct logmasktp {
  u8 data[LOGMASKN],mask[LOGMASKN];
  int debugval;
  char text[LOGMASKTXT];
  struct logmasktp *next;
} logmasktp;


typedef struct   {
  unsigned long curr_wr;
  unsigned long rd;
  unsigned long wr;
  unsigned long wr_atr;
  unsigned long reset_loop;
  unsigned long reset_off;
}  Ser_Delay;

enum { PAR_EVEN=2, PAR_ODD=1, PAR_NONE=0 };
enum { BYTE_DIRECT=0, BYTE_INVERSE=1 };

typedef struct {
  int baud;
  byte size;
  byte parity;
  byte stopbit;
  byte byteconvention;
} Protocol;

typedef struct crypto_system {
  byte *system ;
  /*  byte parity;
  byte byteconvention;
  int baud; */
  Protocol protocol;
  short port;
  char *unixdev;
  byte *atr;
  byte atrsize;
  logmasktp *logmaskp;
  loginfo  mlog,ulog;
  char *filename;
  /* at end (not initialized in main!) : */
  byte extra1,extra2;
  Protocol main_proto;
  byte *main_atr;
  byte main_atrsize;
} cs;

typedef struct filters {
  /* u8 system;*/
  u8 bufsz;
  u8 buf[FILTER_LENGTH+1];
  u8 flags;
  struct filters *next;
} filters;

extern char *SYS_text_abrv[];
extern char *SYS_text_long[];

extern cs *curr_CS, theCS[Sys_Systems];
extern Ser_Delay serial_delay;
extern filters *updfilter[];

#endif



