/* config.c    functions to read config.dat
    Copyright (C) 2000 The Linux Card Emulator Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    http://www.fsf.org/
*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>
#include "config.h"
#include "confutil.h"
#include "main.h"
#include "debug.h"
#include "emu_options.h"

/* local variables for readConfig() */

static int in_section= -1;  /* -2, -1,    0..N-1 */
static int Ldef_system;
static cs *LcsConfig;
static int LcsSize;

Ser_Delay  serial_delay;

char * skip_separator(char *cp) ;

static logmasktp *new_logmask(logmasktp **lm) {
  logmasktp *newlm= (logmasktp *) malloc(sizeof(logmasktp));
  newlm->next=NULL;
  if (! *lm) *lm=newlm;
  else { 
    logmasktp *l;
    for (l=*lm; l->next; l=l->next) ; 
    l->next = newlm;
  }
  return newlm;
}

logmasktp *find_logmask(logmasktp **lm, u8 *ins) {
  logmasktp *l=*lm;
  while (l) {
    u8 *s,*data,*mask;
    int cnt;
    s= ins; 
    data= l->data;
    mask= l->mask;
        for (cnt=0; cnt<LOGMASKN; cnt++) {
      if ( (*s++ & *mask++) != *data++) break;
    }
    if (cnt==5) return l;
    l= l->next;
  }
  return NULL;
}

char * logmask_debug_string(logmasktp *lp) {
  static char strbuf[255];
  if (lp) 
    sprintf(strbuf,"logmask= %x %02x%02x%02x%02x%02x %02x%02x%02x%02x%02x %s [%p->%p]",
	    lp->debugval,
	    lp->data[0],lp->data[1],lp->data[2],lp->data[3],lp->data[4],
	    lp->mask[0],lp->mask[1],lp->mask[2],lp->mask[3],lp->mask[4],
	    lp->text, lp,lp->next);
  else
    strcpy(strbuf,"logmask= NULL");
  return strbuf;
}

void set_protocol_data(Protocol *proto, char *val, char **endptr) {
  byte bconv=0xff;
  byte stp=0xff;
  byte par=0xff;
  unsigned long tmpulval=0;
  char *cp,*cpe;
  if (*val=='-') *endptr=val;
  else tmpulval= strtoul(val,endptr,10);
  cp= *endptr; cpe=0;
  if (cp>val) {
    proto->baud  = (int) tmpulval;
    cp=skip_separator(cp);
    tmpulval = strtoul(cp,&cpe,10);
    /* dummy skip 6/7/8 field */ ;
    if (tmpulval>=6 && tmpulval<=8)
      proto->size=tmpulval; 
    else cpe=cp;
  } else cp=cpe;
  if (cpe>cp) {
    cp=skip_separator(cpe);
    if (tolower(*cp)=='e') par=2;
    else if (tolower(*cp)=='o') par=1;
    else if (tolower(*cp)=='n') par=0;
    else par=0xff;
    if (par <= 2) { 
      cpe=skip_chars(cp,"evenodd");
      proto->parity=par;
    } else cpe=cp;
  }
  if (cpe>cp) {
    cp=skip_separator(cpe);
    if (tolower(*cp)=='1') stp=1;
    else if (tolower(*cp)=='2') stp=2;
    else stp=0xff;
    /* skip stopbit value for now */
    if (stp <= 2) {
      proto->stopbit=stp;
      cpe=skip_chars(cp,"0123456789");
    } else cpe=cp;
  }
  if (cpe>cp) {
    cp=skip_separator(cpe);
    if (tolower(*cp)=='d') bconv=BYTE_DIRECT;
    else if (tolower(*cp)=='i') bconv=BYTE_INVERSE;
    else bconv=0xff;
    if (bconv < 2) {
      cpe=skip_chars(cp,"directnvs");
      proto->byteconvention=bconv;
    } else cpe=cp;
  } else val= *endptr;
  /*
    msgprint(" Baud: %d,%d,%s,%d,%s\n",   proto->baud, proto->size,
    (proto->parity)? (proto->parity&1? "Odd" : "Even"): "None",
    proto->stopbit,(proto->byteconvention) ? "Inverse" : "Direct" );
  */
}

int  ConfigSection(char *section, short system) {
  int i;
  for(i=0; i<LcsSize; i++)
    if(!StrNoCaseCmp(LcsConfig[i].system, section)) {
      in_section = i;
      break;
    }
  if (in_section<0) {
    msgprint("unknown section: [%s]\n", section);
    return 1;
  }
  return 0;
}

int  ConfigVarVal(char *command, char *val, short system) {
  int errcnt=0;
  unsigned long tmpulval;
  char *endptr=val; /* return error if endptr<=val */
  int r;

  if (in_section<0) {
    if (!StrNoCaseCmp(command,"datadir") ) {
      if (!dir_check(&data_dir,val)) errcnt++;
      else endptr=val+1;
    } else if (!StrNoCaseCmp(command,"logdir") ) {
      if (!dir_check(&log_dir,val)) errcnt++;
      else endptr=val+1;
    } else if (!StrNoCaseCmp(command,"delay_read")) {
      serial_delay.rd=strtoul(val,&endptr,0);
    } else if (!StrNoCaseCmp(command,"delay_write")) {
      serial_delay.wr=strtoul(val,&endptr,0);
    } else if (! (StrNoCaseCmp(command,"delay_reset")
		  && StrNoCaseCmp(command,"wr_atr")) ) {
      serial_delay.wr_atr=strtoul(val,&endptr,0);
    } else if (!StrNoCaseCmp(command,"reset_loop")) {
      serial_delay.reset_loop=strtoul(val,&endptr,0);
    } else if (!StrNoCaseCmp(command,"reset_off")) {
      serial_delay.reset_off=strtoul(val,&endptr,0);
    } else
    if (!StrNoCaseCmp(command,"defaultmode")) {
      tmpulval=strtoul(val,&endptr,10);
      if (endptr>val && tmpulval<Sys_Systems)  Ldef_system = (int) tmpulval;
    } else if (!StrNoCaseCmp(command,"colorlog")) {
      set_screen_colors(&endptr,"LOG");
    } else if (!StrNoCaseCmp(command,"colorstat")) {
      set_screen_colors(&endptr,"STAT");
    } else if (!StrNoCaseCmp(command,"colorcomm")) {
      set_screen_colors(&endptr,"COMM");
    } else if (!StrNoCaseCmp(command,"sahuntinc")) {
      tmpulval=strtoul(val,&endptr,0);
  //    if (endptr>val)  sahuntinc =  tmpulval;
    } else if (!StrNoCaseCmp(command,"sahunttimer")) {
      tmpulval=strtoul(val,&endptr,0);
 //     if (endptr>val) sahunttimer = tmpulval;
    } else if (!StrNoCaseCmp(command,"debug")) {
      r=sscanf(val,"%x %x %x %x %x",&(mainlog->debugval),
	     &stddebugval[0],&stddebugval[1],&stddebugval[2],&stddebugval[3]);
      endptr=  r<1 ? val : val+r;      
    } else if (!StrNoCaseCmp(command,"serdebug")) {
      r=sscanf(val,"%x %x %x %x %x",&serdebug,
	     &serdebugval[0],&serdebugval[1],&serdebugval[2],&serdebugval[3]);
      endptr=  r<1 ? val : val+r;      
    } else if (!StrNoCaseCmp(command,"autocheck")) {
      if (!SetU(val, &auto_check, 4, 0, 0, 0x7fffffff)) auto_check=0;
      else endptr = val+1;
    } else if (!StrNoCaseCmp(command,"autosave")) {
      if (!SetU(val, &auto_save, 4, 0, 0, 0x7fffffff)) auto_save=0;
      else endptr = val+1;
    } else ++errcnt; /* error unknown command */
  } else /* in_section */ {
    tmpulval = strtoul(val,&endptr,10);
    if (!StrNoCaseCmp(command,"direct")) {
      LcsConfig[in_section].protocol.byteconvention=BYTE_DIRECT;
      endptr=val+1;
    } else if (!StrNoCaseCmp(command,"inverse")) {
      LcsConfig[in_section].protocol.byteconvention=BYTE_INVERSE;
      endptr=val+1;
    } else if (!StrNoCaseCmp(command,"port")) {
      if (tmpulval<=6) LcsConfig[in_section].port = (int) tmpulval;
      else endptr=val;
    } else if (!StrNoCaseCmp(command,"baud")) {
      if (*val=='-') endptr=val;
      else {
	LcsConfig[in_section].main_proto.baud = 
	  LcsConfig[in_section].protocol.baud = (int) tmpulval;
      }
      if (endptr>val) {
	char *cp,*cpe;
	cp=skip_separator(endptr);
	tmpulval = strtoul(cp,&cpe,10);
	if (endptr>cpe) {
	  LcsConfig[in_section].main_proto.baud = (int) tmpulval;
	}
      }
    } else if (!StrNoCaseCmp(command,"parity")) {
      endptr=val+1;
      if (!StrNoCaseCmp(val,"even")) LcsConfig[in_section].protocol.parity=2;
      else if (!StrNoCaseCmp(val,"odd")) LcsConfig[in_section].protocol.parity=1;
      else if (!StrNoCaseCmp(val,"none")) LcsConfig[in_section].protocol.parity=0;
      else endptr=val;
    } else if(!StrNoCaseCmp(command,"unixdev")) {
      char unix_dev[32];
      char *valstr;
      valstr=SetString_notrsp(val);
      endptr=valstr;
      if(sscanf(valstr,"%31[^ \t]",unix_dev)==1) 
	if (strlen(unix_dev)>0) {
	  /* NB not deallocated after use !!! : */
	  LcsConfig[in_section].unixdev=calloc(strlen(unix_dev)+1,sizeof(char));
	  if (LcsConfig[in_section].unixdev)
	    strcpy(LcsConfig[in_section].unixdev, unix_dev);
	  endptr=val+strlen(val);
	}
    } else if (!StrNoCaseCmp(command,"mainlog") ) {
      char *lfname= SetString_notrsp(val);
      if (strlen(lfname)>0) {
	set_logfilename(&LcsConfig[in_section].mlog, lfname);
	endptr=val+strlen(val);
      }
    } else if (!StrNoCaseCmp(command,"updatelog") ) {
      char *lfname= SetString_notrsp(val);
      if (strlen(lfname)>0) {
	set_logfilename(&LcsConfig[in_section].ulog, lfname);
	endptr=val+strlen(val);
      }
    } else if (!StrNoCaseCmp(command,"main_protocol") ) {
      set_protocol_data(&LcsConfig[in_section].main_proto,val,&endptr);
    } else if (!StrNoCaseCmp(command,"atr_protocol") ) {
      set_protocol_data(&LcsConfig[in_section].protocol,val,&endptr);
    } else if (!StrNoCaseCmp(command,"main_atr") ) {
      /* Read user defined ATR */
      int atrsz=0,minatrsz=8;
      u8 tmpatr[32];
      atrsz=SetU8(val,&tmpatr[0],32,16,0,255);
      if (atrsz>1) {
	/* NB the allocated ATR is not deallocated  !!! : */
	if (atrsz>minatrsz) minatrsz=atrsz;
	LcsConfig[in_section].main_atr=calloc(minatrsz,sizeof(u8));
	memcpy(LcsConfig[in_section].atr,tmpatr,atrsz);
	LcsConfig[in_section].main_atrsize=atrsz;
	endptr=SetU8_valp;
      } else
	endptr= val ;
    } else if (!StrNoCaseCmp(command,"atr") ) {
      /* Read user defined ATR */
      int atrsz=0,minatrsz=8;
      u8 tmpatr[32];
      atrsz=SetU8(val,&tmpatr[0],32,16,0,255);
      if (atrsz>1) {
	/* NB the allocated ATR is not deallocated  !!! : */
	if (atrsz>minatrsz) minatrsz=atrsz;
	LcsConfig[in_section].atr=calloc(minatrsz,sizeof(u8));
	memcpy(LcsConfig[in_section].atr,tmpatr,atrsz);
	LcsConfig[in_section].atrsize=atrsz;
	endptr=SetU8_valp;
      } else
      endptr= val ;
    } else if (!StrNoCaseCmp(command,"logmask") ) {
      int res;
      unsigned long dbval;
      u8 data[6],mask[6];
      char *cp=val;
      cp=skip_white_space(cp);
      dbval=strtoul(val,&endptr,0);
      if ( (res= endptr>val) ) {
	cp=skip_white_space(endptr);
	res = SetCHMASK8(&cp, data, mask, LOGMASKN*2) ;
      }
      if (res) {
	int  cnt;
	logmasktp *lp;
	cp=skip_white_space(cp);
	lp=new_logmask( & LcsConfig[in_section].logmaskp);
	for (cnt=0; cnt<LOGMASKN; cnt++) {
	  lp->data[cnt]= data[cnt];
	  lp->mask[cnt]= mask[cnt];
	}
	lp->debugval= 0xfff & (int) dbval ;
	strncpy(lp->text,cp,LOGMASKTXT); 
	lp->text[LOGMASKTXT-1]='\0';
	/*printf("%s\n",logmask_debug_string(lp));
	subch.name[0]='\0'; subch.name[31]='\0'; subch.flag=0;
	scanf(cp, "%31[^;#\n\r]", subch.name);
	if (*subch.name != '\0') {
	;
	}
	endptr=cp+strlen(cp);*/
      }
    }
  }
  return !(endptr > val);
}


int readConfig(char *file, int *def_system, cs *csConfig, int csSize) {
  int res=0;
  in_section=-1;
  Ldef_system= *def_system;
  LcsConfig=csConfig;
  LcsSize=csSize;
  /* initialize serial delay variables (serfunc.c+) */
  serial_delay.rd=24000;
  serial_delay.wr=12000;
  serial_delay.wr_atr=10000;
  serial_delay.reset_loop=0;
  serial_delay.reset_off=0;
  /* initialiaze color_XXX variables (debug.c+) */


  /* system=0 not used */
  res=ReadConfigFile(file, &ConfigSection, &ConfigVarVal, 0 ); 
  csConfig= LcsConfig;
  *def_system=Ldef_system;
  serial_delay.curr_wr= serial_delay.wr;

#if 0
  /* debug output for logmask data */
  for (in_section=0; in_section<Sys_Systems; in_section++) {
    logmasktp *l;
    printf("LogMask for system=%s:\n",csConfig[in_section].system);
    for (l=csConfig[in_section].logmaskp; l; l= l->next) {
      printf("%s\n",logmask_debug_string(l));    
    }
  }
#endif

  return res;
}

void initConfig_all(cs *csConfig) {
  int sys;
  for (sys=0; sys<Sys_Systems; sys++) {
    csConfig[sys].main_proto.baud=csConfig[sys].protocol.baud;
  }  
  for (sys=0; sys<Sys_ExtraSystems; sys++) {
    updfilter[sys]=NULL;
    /*    updfilter[sys].bufsz=0;
	  updfilter[sys].next=NULL;     */
  }  
}


/* Some tables with constants, from iso7816 */
static int F_FI[16]=    /* etu, clock cycles per transmitted bit */
{ 0,372,558,774,1116,1488,1860,-1,-1,512,768,1024,1536,2048,-1,-1 };
static int F_fs_max[16]= /* in 1kHz units */
{ 0,5000,6000,8000,12000,16000,20000,0,0,5000,7500,10000,15000,20000,-1,-1};
static int D_adjust[16] = /* Bit rate afjustment factor D  in 1/64 units */
{ 0, 64,128,256, 512,1024,   0, 0, 0,  0, 32,  16,   8,  4,  2,  1};
static byte I_II[4]= { 25, 50, 100, 0}; /* max current I */

static int ATRDEBUG=1, atrdebug=0;

static byte nextatr(byte **p1, byte **p2, char *txt) {
  byte b;
  b = *(*p2)++;
  if (p1) *(*p1)++ = b;
  if (atrdebug) msgprint(" %s%02x",txt,b);
  return b;
}

void build_ATR(cs *currCS, int c_sys, int deb_sys) {
  byte *atr, **atrnewp, atrXi;
  byte *pTA1=NULL;
  byte useTXi, historical;
  byte TS,T0,FI=1,DI=1,II=0,PI1=0,PI2=0, N=0;
  int F,D,I,P,fs,tmp;
  float f,etu;
  byte adjustAtrSize = 1;

  atr= currCS->atr;
  atrnewp= NULL;
  atrdebug= (ATRDEBUG && deb_sys>=0 && deb_sys==c_sys) ;
  if (atrdebug) msgprint("Build/check ATR(%s) :\n",currCS->system);

  /* TS */
  /* write updated TS:direct/inverse to  currCS->atr == *atr */
  *atr=(currCS->protocol.byteconvention) ? 0x3f : 0x3b;
  TS=nextatr(atrnewp,&atr,"ts:");

  /* T0 */
  T0= nextatr(atrnewp,&atr,"t0:");
  useTXi= T0 & 0xf0;
  historical= T0 & 0x0f;

  atrXi=0;
  while (useTXi && (atr-currCS->atr)<currCS->atrsize) {
    byte tmp;
    char s[10];
    atrXi++;
    sprintf(s,"tA%d:",atrXi);
    /* TAi - TDi*/ 
    if (useTXi & 0x10) {
      s[1]='a';
      tmp= nextatr(atrnewp,&atr,s);
      if (atrXi==1) {
	pTA1= atr-1;
	FI= tmp >> 4;
	DI= tmp &0xf;	
      }
    }
    if (useTXi & 0x20) {
      s[1]='b';
      tmp= nextatr(atrnewp,&atr,s);
      if (atrXi==1) {
	II= (tmp >> 5) & 0x3;
	PI1= tmp &0x1f;
      } else if (atrXi==2) PI2= tmp &0xff;
    } 
    if (useTXi & 0x40) {
      s[1]='c';
      tmp= nextatr(atrnewp,&atr,s);
      if (atrXi==1) N= tmp;
    }
    if (useTXi & 0x80) {
      s[1]='d';
      tmp= nextatr(atrnewp,&atr,s);
      if (atrXi==1) {
      }
      useTXi= tmp &0xf0;
      if((tmp & 0xf) == 0xe)
	adjustAtrSize = 0;
      /* atrXi==1: bit 0..3 are protocol type , default 0*/
    } else useTXi=0;
  }
  if (atrdebug) msgprint(" hist.(%d):",historical);
  for (atrXi=0; atrXi<historical; atrXi++) nextatr(atrnewp,&atr,"");
  if (atrdebug) msgprint("\n");

  tmp= (atr- currCS->atr);
  if (currCS->atrsize != tmp) {
    int i;
    byte *p;
    msgprint("*** ATR(%d):", currCS->atrsize);
    p=currCS->atr;
    for (i=currCS->atrsize; i>0; i--) msgprint(" %02x",*p++);
    msgprint("\n");
    /* adjust atrsize if nescesary */
    if (currCS->atrsize!=tmp && adjustAtrSize) {
      currCS->atrsize= tmp; 
      msgprint("*** Warning : ATR(%s) size mismatch, size=%d, should be=%d ***\n",
	       currCS->system,currCS->atrsize,tmp);
    }
  }

  atr=currCS->atr;
  if (atrdebug) {
    msgprint(" fi=%x di=%x ii=%x pi1=%x pi2=%x, guardtime N=%d\n",
	     FI,DI,II,PI1,PI2,N);
  }
  F=F_FI[FI]; fs= F_fs_max[FI];
  if (fs==0) fs=3571;
  if (F==0) F=372;
  D= D_adjust[DI];
  if (D==0) D=64;
  I=I_II[II];
  P=PI1<5||PI1>25?0:PI1;
  if (atrdebug) {
    msgprint("%s clk., max fs=%6.3fMHz, F=%d, D=%4.2f, max I=%dmA, Vpp=%dV\n",
	     FI==0? "Internal" : "External", fs/1000.0,F,D/64.0,I,P);
  }

  if (FI==0) etu=1/9600.0/(D/64.0)*1000;
  else etu=F*1.0/(D/64.0)/fs;
  f=1/etu*1000; /* bps */
  etu=etu * 1000*1000; /* micro seconds s */
  if (atrdebug) {
    msgprint("Work @ %6.3fMHz %6.3fMHz %6.3fMHz\n",fs/1000.0,3.5712,4.0);
    msgprint(" etu   %6.0fs  %6.0fs  %6.0fs\n", 
	     etu,etu*fs/3571.2,etu*fs/4000);
    msgprint(" bps %8.0f, %8.0f  %8.0f [bits per second]\n",
	     f,f*3571.2/fs,f*4000/fs);
    msgprint("ATR(%s): %s, BPS=%d, parity=%s\n",
	     currCS->system,currCS->protocol.byteconvention? "inverse":"direct",
	     currCS->protocol.baud, currCS->protocol.parity?(currCS->protocol.parity==1?"odd":"even"):"odd");
  }
}

void read_addfilter(char *val, short sys, byte flags) {
  filters *newf= (filters *) malloc(sizeof(filters));
  newf->bufsz=0;
  newf->next=NULL;
  if (updfilter[sys] == NULL) {
    updfilter[sys]=newf;
  } else {
    filters *f=updfilter[sys];
    while ( f->next ) { 
      f=f->next;
    }
    f->next=newf;
  }
  newf->bufsz=SetU8(val,newf->buf,FILTER_LENGTH,16,0,255);
 // newf->buf[newf->bufsz] = IN_USE;
  newf->flags |= flags;
}

