#ifndef CONFUTIL_H
#define CONFUTIL_H
/* configutil.h    functions to read configuration files
    Copyright (C) 2000 The Linux Card Emulator Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    http://www.fsf.org/
*/
#include <stdio.h>
#include "common.h"

typedef int (*ProcSection) (char *s, short system) ;
typedef int (*ProcVariable) (char *var, char *val, short system) ;

typedef struct comment_rec {
  struct comment_rec *next;
  char * text;
} comment_rec;

enum { MM2base=17, Nordic4base=18 };

enum { LINESIZE=1024, COMMANDSZ=32 };

int file_exist_rd(char *fname) ;
int dir_exist(char *dirname, int *err) ;
int dir_check(char **dir, char *val) ;
char *alloc_cat_dir_file(char *dir, char *filename) ;
int dev_ok(char *fname);
char * find_rc_file(char *rc_name, char *rc_home_name) ;

extern int config_lineno;

int ReadConfigFile(char *fname, ProcSection sect, ProcVariable pvar,
		   short system);


void add_comment(comment_rec **first, char *val);
void print_comments(FILE *keyfile, comment_rec *comment_ptr) ;

void keys_to_string(char *s, byte n, byte keys[]);

int SetUL(char *val,unsigned long  *var, unsigned sz, int base);
int SetU (char *val,unsigned  *var, unsigned sz, int base,
	  unsigned  min, unsigned max);
int SetU8(char *val, u8 *var, unsigned sz, int base, int min, int max) ;
int SetU32(char *val, u32 *var, u32 *pvalid, unsigned sz, int base, u32 min, u32 max) ;
int SetCHMASK(char **val, u32 *var,u32 *mask,unsigned sz) ;
int SetCHMASK8(char **val, u8 *var,u8 *mask,unsigned sz) ;
char *SetString(char *val) ;
char *SetString_notrsp(char *val) ;

int SetDateTime(char *val, datetime *dt);

void str_eatspaces(char *p) ;
int StrNoCaseCmp(const char *s1, const char *s2);
int StrNoCaseCmpN(const char *s1, const char *s2, const unsigned int n);

unsigned HextoMM2(u8 H, unsigned pos) ;
u8 MM2toHex(unsigned mm2, unsigned *pos, short key_length) ;

char * skip_white_space(char *cp) ;
char * skip_chars(char *cp, char *charset) ;
extern char *HextoChar;
extern char *SetU8_valp;

void Nordic4toHex(u8 *hx, u8 *id, u8 *kix, u16 *n4) ;
void HextoNordic4(u16 *n4, u8 *id, u8 *keys, byte keyindex) ;
extern u8 nord4_ID[3], nord4_key;


#endif






