/*  emuoption.c - get command line options, + misc ncurses functions
    (C) 2002-2004 The Linux Card Emulator Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    http://www.fsf.org/ */

/* global variable */

#ifdef USE_WIN32
#include <windows.h>
#else
#include <termios.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/stat.h>
#endif

#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "common.h"
#include "confutil.h"
#include "emu_options.h"
#include "debug.h"

#ifdef USE_READLINE
#include <readline/readline.h>
#ifndef _WIN32
#define USE_HISTORY
#include <readline/history.h>
#endif
#endif


comm_options_t comm_opt = {0, 0, -1, 0, 0, -1, {-1,-1,-1}, 0, 0, 0, 0, 0 ,0 } ;


static char *optionHelp[] =
{ 
#ifdef USE_NCURSES 
  "usage: lincardemu [options] strings ...",
  "   -c   use ncurses",
  "   -o   use stdout",
  "   -d/dev/tty5  use /dev/tty5",
  "   -wN[slc]* set command window size=N, and optional sequence of windows",
#endif
  "   optimalizovane pre irdeto 3, podpora CkeyCh_v1",
  "   -h   this help",
  
  NULL
};

char *parse_command_line(int argc, char **argv) {
  char *cp;
  int opt;
  
#ifdef USE_HISTORY
  using_history();
#endif
  while (0<=(opt= getopt(argc,argv,"Vocrl:hs:w:d:f:"))) {
    switch(opt) {
    case 'o':
    case 'c': 
      comm_opt.curses= opt=='c';
      break;
    case 'r': 
      comm_opt.data_files_RO= 1;
      break;
    case 'l':
      {
	short logtype = (short) strtol(optarg,&cp,0);
	if (cp==optarg) {
	  comm_opt.flags|=OPT_ERROR;
	  printf("#%s: error in option -s%s\n",argv[0],optarg);
	  comm_opt.system=-1;
	} else {
	  comm_opt.logcard = 0;
	  comm_opt.blind_logcard = 0;
	  if (logtype & 0x18) 
	    /* bit 3-4 */
	    comm_opt.blind_logcard = (logtype >> 3) & 0x3;
	  else if (logtype & 0x7)
	    /* bit 0-2 */
	    comm_opt.logcard = logtype;	   
	}
      }
      break;
    case 'h':
      comm_opt.flags |= OPT_HELP;
      break;
    case 's':
      comm_opt.system= (short) strtol(optarg,&cp,0);
      if (cp==optarg|| comm_opt.system>=Sys_Systems) {
	comm_opt.flags|=OPT_ERROR;
	printf("#%s: error in option -s%s\n",argv[0],optarg);
	comm_opt.system=-1;
      } 
      break;
    case 'f':
      comm_opt.cfg_file= optarg;
      fprintf(stderr,"config file='%s'\n",optarg);
      break;
    case 'd':
      comm_opt.dev= optarg;
      {
	int r;
        r=dev_ok(comm_opt.dev);
	fprintf(stderr,"device=%s res=%d\n",optarg,r);
	if ( !r  )  comm_opt.flags |= OPT_ERRDEV|OPT_ERROR; 
      }
      break;
    case 'w': 
      {
	char *seqstr="lsc ,";
	comm_opt.commlines= (short) strtol(optarg,&cp,0);
	if (cp==optarg) {
	  comm_opt.flags|=OPT_ERROR;
	  printf("#%s: error in option -s%s\n",argv[0],optarg);
	  comm_opt.commlines=-1;
	} else if (strchr(seqstr,tolower(*cp))){
	  int cnt;
	  char *sq;
	  for (cnt=0; cnt<WINCOLOR_MAX; cnt++) comm_opt.conf_line[cnt]=cnt;
	  while ((sq=strchr(seqstr,tolower(*cp)))) {
	    if ((sq-seqstr)<3) comm_opt.conf_line[sq-seqstr]= cnt++;
	    cp++;
	  }
	} else if (*cp) {
	  comm_opt.flags|=OPT_ERROR;
	  printf("#%s: error in option -s%s\n",argv[0],optarg);
	}
      }
      break;
    case '?':
      printf("#%s: invalid option -- %c\n",argv[0],opt);
      comm_opt.flags |= OPT_ERROR;
      break;
    case ':':
      printf("#%s: option requires an argument -- %c\n",argv[0],optopt);
      /*printf(" missing parameter\n");*/
      comm_opt.flags |= OPT_ERROR;
      break;
    case 'V':
      printf("Version: %s\n",decoder_name);
      /*printf(" missing parameter\n");*/
      comm_opt.flags |= OPT_ERROR;
      break;
    default :
      printf(" ?? getopt returned character code 0x%x ??\n",opt);
      comm_opt.flags |= OPT_ERROR;
      break;
    }
  }
  comm_opt.argv=argv;
  comm_opt.argc=argc;
  comm_opt.argcp=argc;
  comm_opt.argcp=optind;
  comm_opt.argvp=NULL;
  if (comm_opt.argcp< argc){
    comm_opt.argvp= argv[comm_opt.argcp++];
    printf("commands argv=%s\n",comm_opt.argvp);

#if 0
    printf ("non-option ARGV-elements: ");
    while (optind < argc)
      printf ("\"%s\" ", argv[optind++]);
    printf ("\n");
#endif
  }
  if (comm_opt.flags & (OPT_ERROR|OPT_HELP|OPT_ERRDEV)) {
    char **emsg;
    emsg= optionHelp;
    while (*emsg) {
      scrprint("%s\n",*emsg);	  
      emsg++;
    }
    return NULL;
  }
  return "";
}

int console_getch() {
  /* if still parameters (non-option) from commandline send them
     as keystrokes before calling scrgetch() 
  */
  int c;
  if (comm_opt.argvp==NULL) return scrgetch();
  /* accept null-chars / empty string... from command line arguments: */
  c= *(comm_opt.argvp);
  if (c) comm_opt.argvp++;
  if (! *comm_opt.argvp) {
    if (comm_opt.argcp<comm_opt.argc)
      comm_opt.argvp= comm_opt.argv[comm_opt.argcp++];    
    else comm_opt.argvp=NULL;
  }
  return c;
}



static char *console_getch_argv() {
  char *cp;
  if (comm_opt.argvp==NULL) return NULL;
  /* accept null-chars / empty string: */
  cp= comm_opt.argvp;
  if (comm_opt.argcp<comm_opt.argc)
    comm_opt.argvp= comm_opt.argv[comm_opt.argcp++];    
  else comm_opt.argvp=NULL;
  return cp;

}

/*
  int console_confirm (char *prompt) {
  return 1;
  }
*/

char * console_getline(char *prompt) {
  static char lnbuf[128];
  char *ln=NULL;
  ln= console_getch_argv(); 
  lnbuf[0]='\0';
  if (ln) scrprint("%s %s\n",prompt,ln);
#ifdef USE_NCURSES
  else if (commwin) {
    echo(); nocbreak();
    scrprint("%s",prompt);
    wclrtoeol(commwin);
    wrefresh(commwin);
    wgetnstr(commwin,lnbuf,127);   
    lnbuf[127]='\0';
    ln=lnbuf;
    noecho(); cbreak();
  } 
#endif
  else {
    unsigned char mout=0,uout=0;
    /* disable output to screen 
       read a line 
       and reenable log output to screen */
    mout= mainlog->out;
    uout= updlog->out;
    mainlog->out &= ~To_console;
    updlog->out &= ~To_console;
    tty_inputmode(1);
#ifdef USE_READLINE
    ln = readline(prompt);
    if (ln) {
      char *cp;
      for (cp=ln; isspace(*cp); cp++) ;
      if (*cp) {
#ifdef USE_HISTORY
	add_history(cp);
#endif
	strncpy(lnbuf,cp,128);
	lnbuf[127]='\0';
      }
      free(ln); /* free readline allocated string */
    }
    ln=lnbuf;
#else
    /* no readline, simple read string.... */
    scrprint("%s",prompt);
#ifdef USE_WIN32
    if (!fgets(lnbuf,128,stdin)) lnbuf[0]='\0';
#else
    if (!fgets(lnbuf,128,stdin)) lnbuf[0]='\0';
#endif
    if (strlen(lnbuf)>0) {
      ln=lnbuf+strlen(lnbuf)-1;
      if ( *ln != '\n' && *ln != '\r') {
	/* skip to end of line */
	int c;
	while ( (c=getc(stdin))!=EOF)
	  if (c=='\n'||c=='\r') break;
	/*else scrprint("skip:%02x ",c);*/
      } else *ln='\0';
    }
    for (ln=lnbuf; isspace(*ln); ln++) ;
    /*scrprint("END= %02x\n",lnbuf[strlen(lnbuf)-1]);*/
#endif
    tty_inputmode(0);
    mainlog->out=mout;
    updlog->out=uout;
  }
  /* log input : */
  mlogprint(0x1|DEB_ONLYLOG,"%s %s\n",prompt,ln);
  return ln;
}



color_opt_t color_log, color_stat, color_comm, win_colors[WINCOLOR_MAX];

#ifndef USE_NCURSES
void wincolors_init() { }

#else

static int conflines[WINCOLOR_MAX];

/* compare sequence (for qsort) */
static int wincolor_seq_cmp(const void *c1, const void *c2) {
  int l1= win_colors[*((int *) c1)].conf_line;
  int l2= win_colors[*((int *) c2)].conf_line;
  return l1< l2 ? -1 : l1==l2 ? 0 : 1;
}

static void adjust_winsizes_setcolors(int lines) {
  int i,sum;

  if (comm_opt.commlines>0) win_colors[WINCOLOR_COMM].lines=comm_opt.commlines;
  for (i=0; i<WINCOLOR_MAX; i++) 
    if (comm_opt.conf_line[i]>=0) win_colors[i].conf_line=comm_opt.conf_line[i];
  win_colors[WINCOLOR_STAT].lines= 1; /* always = 1 */
  i=LINES/3;
  if (i<1) i=1;
  else if (i>10) i=10;
  /* check window size :  (at least 1 line per window !) */
  if (win_colors[WINCOLOR_COMM].lines>LINES-WINCOLOR_MAX)
    win_colors[WINCOLOR_COMM].lines=i;
  else if (win_colors[WINCOLOR_COMM].lines<1)
    win_colors[WINCOLOR_COMM].lines=i;
  if (win_colors[WINCOLOR_LOG].lines>LINES-WINCOLOR_MAX)
    win_colors[WINCOLOR_LOG].lines=LINES-i-1;
  else if (win_colors[WINCOLOR_LOG].lines<1)
    win_colors[WINCOLOR_LOG].lines=LINES-i-1;
  
  /* now all values in range 1..LINES-2 ; check sum : */
  sum=win_colors[WINCOLOR_LOG].lines+win_colors[WINCOLOR_COMM].lines+1;
  if (sum<LINES) {
    /* expand LOG window to fill the rest */
    win_colors[WINCOLOR_LOG].lines += LINES-sum;
  } else if (sum>LINES) {
    /* sum to big, adjust in ratio 1:3 */
    win_colors[WINCOLOR_COMM].lines -= (sum-LINES)/3;
    win_colors[WINCOLOR_LOG].lines= LINES-win_colors[WINCOLOR_COMM].lines-1;
  }

  /* sort to correct sequence (sort key : win_colors[].conf_lines) */
  for(i=0; i<WINCOLOR_MAX; i++) conflines[i]=i;
  qsort(conflines, WINCOLOR_MAX, sizeof(int), wincolor_seq_cmp );

  win_colors[conflines[0]].y=0;
  win_colors[conflines[1]].y=win_colors[conflines[0]].lines;
  win_colors[conflines[2]].y=
    win_colors[conflines[1]].y+win_colors[conflines[1]].lines;

#if 0
  for(i=0; i<WINCOLOR_MAX; i++) {
    color_opt_t *copt = &win_colors[i];
    printw(" win_color[%d] conf_lines=%d lines=%d y=%d  fcol=%d bcol=%d fattr=%x battr=%x\n",
	   i,copt->conf_line,copt->lines,copt->y,
	   copt->fcol,copt->bcol,copt->fattr,copt->battr);
  }
  getch();
#endif
  
  win_colors[WINCOLOR_LOG].win = &logwin;
  win_colors[WINCOLOR_STAT].win= &statwin;
  win_colors[WINCOLOR_COMM].win= &commwin;
  
  for (i=0; i<WINCOLOR_MAX; i++) {
    color_opt_t *c= &win_colors[i];
    *(c->win)= newwin(c->lines,0,c->y,0);
    if ( *(c->win) ) {
      idlok(*(c->win),TRUE);
      scrollok(*(c->win),TRUE);
      init_pair(i+1,c->fcol,c->bcol);
      wbkgd(*(c->win) ,' '|c->battr |COLOR_PAIR(i+1));
#if  NCURSES_VERSION_MAJOR>=5
      wattr_set(*(c->win),c->fattr,i+1,NULL);
#else
      wattr_set(*(c->win),c->fattr);
#endif
    }
  }
}

void wincolors_init() {
  int  i;
  color_opt_t *copt;
  copt= &win_colors[0];
  copt->fcol = copt->bcol = -1;
  copt->fattr= copt->battr =A_NORMAL;
  copt->conf_line=-WINCOLOR_MAX;
  copt->lines=-1;
  for (i=1; i<WINCOLOR_MAX; i++) {
    win_colors[i]= *copt;
    win_colors[i].conf_line= -WINCOLOR_MAX+i;
  }
}
#endif

void setup_inputmode() {
#ifdef USE_WIN32
  
#else
#ifndef USE_NCURSES
  comm_opt.curses=0;
#endif
  if (!comm_opt.curses) {
    tty_inputmode(0);
    /* tty_inputmode(0,fdevout,fdevin); */
#ifdef USE_NCURSES

  } else {
    if (comm_opt.dev) {
      newterm(NULL,comm_opt.fdevout, comm_opt.fdevin);
    }
    initscr(); 
    
    /* if ( LINES<3) .... */
    start_color();
    if (has_colors()) {
      use_default_colors();
    }
    cbreak(); 
    noecho();
    nonl();
    idlok(stdscr,TRUE);
    scrollok(stdscr,TRUE);
    intrflush(stdscr, FALSE);

    adjust_winsizes_setcolors(LINES);

    if (commwin) {
      keypad(commwin, TRUE);
    }
    if (has_colors() && commwin && logwin) {
      wprintw(logwin," LOG WINDOW\n");
      wprintw(statwin," STATUS LINE");
      wprintw(commwin," COMMAND WINDOW\n");
      wrefresh(statwin);wrefresh(logwin);  wrefresh(commwin);
    }
#endif
  }
#endif
}

void restore_inputmode() {
#ifndef USE_WIN32
  if (!commwin) tty_inputmode(1);
#ifdef USE_NCURSES
  else {
    endwin();
    commwin=NULL;
    statwin=NULL;
    logwin=NULL;
    mainlog->out &= ~(To_win|To_commwin);
    mainlog->out |=To_stdout;
  }
#endif
#endif
}



#ifndef USE_NCURSES
static void set_y_colors(char **val, color_opt_t *copt, short height) {
  (*val)++;
}
#else
typedef struct {
  char * name;
  int  value;
} coltab_t;

static coltab_t  coltab[] = {
  {"BLACK", COLOR_BLACK},
  {"RED",COLOR_RED},
  {"GREEN",COLOR_GREEN},
  {"YELLOW",COLOR_YELLOW},
  {"BLUE",COLOR_BLUE},
  {"MAGENTA",COLOR_MAGENTA},
  {"CYAN",COLOR_CYAN},
  {"WHITE",COLOR_WHITE},
  {"DEFAULT",-1},
  {"NORMAL",A_NORMAL},
  {"STANDOUT",A_STANDOUT},
  {"UNDERLINE",A_UNDERLINE},
  {"REVERSE",A_REVERSE},
  {"BLINK",A_BLINK},
  {"DIM",A_DIM},
  {"BOLD",A_BOLD},
  {NULL,0}
};

static int find_coltab(char **cp) {
  char chword[16];
  char *p;
  int  cltn;
  coltab_t *cltp;
  strncpy(chword, *cp, 16);
  chword[15]='\0';
  p= strchr(chword,' ');
  if (p) *p='\0';
  if (strlen(chword)<=0) {
    return -1;
  }
  cltn=0; 
  for ( cltp=coltab; cltp->name && StrNoCaseCmp(chword,cltp->name); cltp++)
    cltn++;
  if (!cltp->name) cltn=-2;
  else *cp = *cp+strlen(chword);
  return cltn;
}
static void set_col_attr(int *col, attr_t *atr, char **cp, char **cpe) {
  int cltn;
  *cp=*cpe;
  do {
    *cpe=skip_white_space(*cpe);
    cltn= find_coltab(cpe);
    if (cltn>8) (*atr) |= coltab[cltn].value;
  } while (cltn>8);
  if (cltn>=0) *col = coltab[cltn].value;
  else if (cltn< -1) *cp=*cpe;
}

static void set_y_colors(char **val, color_opt_t *copt, short height) {
  char *cpe=*val, *cp;
  cpe=skip_white_space(cpe);
  cp= cpe-1;
  copt->conf_line= config_lineno; /* for sequence of windows */
  if (height && isdigit(*cpe)) {
    cp=cpe;
    copt->lines= (short) strtoul(cp,&cpe, 10);
    if (cpe>cp && !isspace(*cpe)) cp=cpe;
  } 
  copt->fcol=copt->bcol=-1;
  copt->fattr= copt->battr= A_NORMAL;
  if (cpe>cp && *cpe) set_col_attr(&copt->fcol,&copt->fattr, &cp, &cpe);
  if (cpe>cp && *cpe) set_col_attr(&copt->bcol,&copt->battr, &cp, &cpe);
  if (cpe > cp ) *val=cpe;  
#if 0
  printf(" y:%d fcol:%d bcol:%d fattr:%x battr:%x %p %p\n",
	 copt->y, copt->fcol, copt->bcol, (int) copt->fattr, (int) copt->battr,
	 *val,cpe);
#endif
}
#endif

void set_screen_colors(char **endptr, char *scrtype) {
  if (!strcmp(scrtype,"LOG")) {
    set_y_colors(endptr,&win_colors[WINCOLOR_LOG],1);
  } else if (!strcmp(scrtype,"STAT")) {
    set_y_colors(endptr,&win_colors[WINCOLOR_STAT],0);      
  } else if (!strcmp(scrtype,"COMM")) {
    set_y_colors(endptr,&win_colors[WINCOLOR_COMM],1);
  }
}

