#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <ctype.h>
#include <time.h>
#include <pthread.h>

#include "common.h"
#include "debug.h"
#include "emu_options.h"
#include "confutil.h"
#include "config.h"
#include "irdeto.h"
#include "newcamd-client.h"

void exit_nicely(int n) ;
char *decoder_name= "Lincardemu for irdeto" ;
static pthread_t serialThread;
static int have_thread=0;
static int got_signal=0;

void sighandler_general(int sv) {
  got_signal=sv;
  msgprint("!! Signal 0x%x received \n",sv);
  if (sv==SIGHUP||sv==SIGINT||sv==SIGQUIT||sv==SIGTERM) {
    msgprint("saving & shutting down\n");
    if (have_thread  ) {
      have_thread=0;
      pthread_kill(serialThread, 0);
    }
    exit_nicely(0);
  }
}

void init_signalhandler() {
  signal(SIGPIPE,sighandler_general);
  signal(SIGHUP,sighandler_general);
  signal(SIGINT,sighandler_general);
  signal(SIGQUIT,sighandler_general);
  signal(SIGTERM,sighandler_general);
 }


enum { Error_Reset=1, Auto_Reset=2, Manual_Reset=4 , Auto2_Reset=8};

static byte do_reset = 0;
void force_reset() { do_reset=Auto2_Reset; }
static char *rc_file=NULL, *rc_file_sys=NULL;

void exit_nicely(int n) {
    flush_logs();
    sleep(2);
    restore_inputmode();
    serial_port_shutdown(0);
    exit(n);
}

void *serial_handling(void *param) 
 {
    byte resb,ins[5];
    int  res;
    unsigned int counter=0;
    do_reset=Auto_Reset;
     
    while(1) 
    {
     if (do_reset && !comm_opt.logcard)
       {
        static int atr_c;
	      mlogprint(0x1,"Waiting for reset\n") ;
        do {
        GetReset();
	       atr_c = sendbyte_n(0x2|comm_opt.logcard,
			   theCS[Sys_Irdeto].atr,theCS[Sys_Irdeto].atrsize);
	      if (atr_c != theCS[Sys_Irdeto].atrsize )
	       mlogprint(0x2,"- error statusd=%d (!=%d)\n",atr_c,theCS[Sys_Irdeto].atrsize);
	      else mlogprint(0x2," ok - ");
	          } while(atr_c != theCS[Sys_Irdeto].atrsize);
	       do_reset=0;
        mlogprint(0x1,"ATR sent successfully\n") ;
       }
      
             
       for(counter=0; counter< 5 ; counter++)
       {
	      res = readbyte(0);
	      ins[counter]= resb = (byte) res &0xff;
	      if(res< 0)
         {
	       msgprint("Error receiving INS[%d] (%d)\n",counter,res);
	       break;
	       }   
        }
      
        if (counter!= 5) do_reset=Error_Reset;
	      DoIrdeto(ins, ins[0], ins[1], ins[2], ins[3], ins[4]);
    }
    return NULL;
  }

static int set_protocol(Protocol *proto, cs *theCSp) {
  int  res=0;
  res=initPort_season(proto->size, proto->parity, proto->stopbit,
	       proto->byteconvention, proto->baud);
  if (res>=0){
    char *udev;
    if (!(udev= theCSp->unixdev)) udev="";
    msgprint("System:%s Protocol: %d %d,%s,%d,%s COM%d,%s\n",
	     theCSp->system, proto->baud, proto->size,
	     (proto->parity)? (proto->parity&1? "Odd" : "Even"): "None",
	     proto->stopbit,(proto->byteconvention) ? "Inverse" : "Direct",
	     theCSp->port,udev);
  }
  return res;
}

int initialize_port(cs *theCSp) {
  int  res=0;
  res= set_protocol(&theCSp->protocol, theCSp);
  return res;
}

static char *CommandHelp[] =
{  " d: toggle debug , 4 different values",
    NULL
};

int main(int argc, char **argv) {
  
  char *commstr;
  int def_system;
  unsigned long *pcurr_delay;
  byte debug = 1;
  int  cfg_errors= -1;
  int command_loop=1;

  set_time_stamp();
  curr_CS = &theCS[Sys_Irdeto]; /* set some sane values  */

  commstr=parse_command_line(argc,argv);
  if (!commstr) exit(18);
  
  set_new_loginfo(&curr_CS->mlog,&curr_CS->ulog);

  rc_file_sys="lincardemu.dat";
  if (! rc_file_sys) {
    msgprint("couldn't find configuration file %s\n","lincardemu.dat");
    exit_nicely(1);
  } else {
    cfg_errors=readConfig(rc_file_sys, &def_system, theCS, (sizeof(theCS)/sizeof(cs)));
    if (cfg_errors) {
      msgprint("..syntax error(s) in config file %s\n",rc_file_sys);
      exit_nicely(1);
    }
  }
  msgprint("Open port\n");
  curr_CS= &theCS[Sys_Irdeto];
  openPort_season(theCS[Sys_Irdeto].port,theCS[Sys_Irdeto].unixdev);
 
  setup_inputmode();
  build_ATR(&theCS[Sys_Irdeto],Sys_Irdeto,Sys_Irdeto);
     
  if (initialize_port(curr_CS)<0) {
    exit_nicely(1);
  } 
  
  cfg_errors=init_newcamd();
  if (cfg_errors!=0)  exit_nicely(16);
  
  if(pthread_create(&serialThread, NULL, serial_handling, NULL) != 0) {
    errprint(errno,"Couldn't spawn serial thread\n");
    exit_nicely(17);
  }
  pthread_detach(serialThread);
  have_thread=1;
  init_signalhandler();

  do {
    int con_ch=console_getch();
    if (isalpha(con_ch)) con_ch=toupper(con_ch);
    switch(con_ch) {
    
    case 'D':
      debug=debug_toggle();
      scrprint("Debugging %s (%d:0x%x)\n",  debug ? "on" : "off", debug,mainlog->debugval);
      break;

    default :
      {
	char **emsg;
	emsg= CommandHelp;
	scrprint("Commands: %02x\n", con_ch);
	while (*emsg) {
	  scrprint("%s\n",*emsg);
	  emsg++;
	}
      }
      break;
    }
  } 
  while(command_loop);
   if (have_thread ) {
    pthread_kill(serialThread, 0);
  }

  exit_nicely(0);

  return 0;
}
